@@ -26,18 +26,19 @@
 #ifndef _AUTH_H_
 #define _AUTH_H_
 
-/**
- * @brief Actions to take on clients.
- */
-typedef enum {
-	AUTH_MAKE_DEAUTHENTICATED = 1, /**< To make client 'deauthenticated' */
-	AUTH_MAKE_AUTHENTICATED = 2 /**< To make client 'authenticated' */
-} t_authaction;
-
-/** @brief Take action on a single client */
-void auth_client_action(const char ip[], const char mac[], t_authaction action);
+int auth_client_deauth(unsigned id, const char *reason);
+int auth_client_auth(unsigned id, const char *reason);
+int auth_client_trust(const char *mac);
+int auth_client_untrust(const char *mac);
+int auth_client_allow(const char *mac);
+int auth_client_unallow(const char *mac);
+int auth_client_block(const char *mac);
+int auth_client_unblock(const char *mac);
 
 /** @brief Periodically check if connections expired */
 void *thread_client_timeout_check(void *arg);
 
+/** @brief Deauth all authenticated clients */
+void auth_client_deauth_all();
+
 #endif
