@@ -24,10 +24,12 @@
     @author Copyright (C) 2007 Paul Kube <nodogsplash@kokoro.ucsd.edu>
 */
 
+#define COPYRIGHT "NodogSplash, Copyright (C) 2016 - 2019, The Nodogsplash Contributors"
+
 #ifndef _CONF_H_
 #define _CONF_H_
 
-#define VERSION "2.0.0-git"
+#define VERSION "4.0.2"
 
 /*@{*/
 /** Defines */
@@ -47,44 +49,39 @@
 #define DEFAULT_CONFIGFILE SYSCONFDIR"/nodogsplash/nodogsplash.conf"
 #endif
 #define DEFAULT_DAEMON 1
-#define DEFAULT_DEBUGLEVEL LOG_WARNING
+#define DEFAULT_DEBUGLEVEL 1
 #define DEFAULT_MAXCLIENTS 20
 #define DEFAULT_GATEWAY_IPRANGE "0.0.0.0/0"
 #define DEFAULT_GATEWAYNAME "NoDogSplash"
 #define DEFAULT_GATEWAYPORT 2050
+#define DEFAULT_FASPORT 0
+#define DEFAULT_FAS_SECURE_ENABLED 1
+#define DEFAULT_FASPATH "/"
 #define DEFAULT_REMOTE_AUTH_PORT 80
-#define DEFAULT_CHECKINTERVAL 60
-#define DEFAULT_CLIENTTIMEOUT 10
-#define DEFAULT_CLIENTFORCEOUT 360
+#define DEFAULT_CHECKINTERVAL 30
+#define DEFAULT_SESSION_TIMEOUT 0
+#define DEFAULT_PREAUTH_IDLE_TIMEOUT 10
+#define DEFAULT_AUTH_IDLE_TIMEOUT 120
 #define DEFAULT_WEBROOT "/etc/nodogsplash/htdocs"
 #define DEFAULT_SPLASHPAGE "splash.html"
-#define DEFAULT_INFOSKELPAGE "infoskel.html"
-#define DEFAULT_IMAGESDIR "images"
-#define DEFAULT_PAGESDIR "pages"
+#define DEFAULT_STATUSPAGE "status.html"
 #define DEFAULT_AUTHDIR "nodogsplash_auth"
 #define DEFAULT_DENYDIR "nodogsplash_deny"
+#define DEFAULT_PREAUTHDIR "nodogsplash_preauth"
 #define DEFAULT_MACMECHANISM MAC_BLOCK
-#define DEFAULT_PASSWORD_AUTH 0
-#define DEFAULT_USERNAME_AUTH 0
-#define DEFAULT_PASSWORD_ATTEMPTS 5
-#define DEFAULT_AUTHENTICATE_IMMEDIATELY 0
 #define DEFAULT_SET_MSS 1
 #define DEFAULT_MSS_VALUE 0
 #define DEFAULT_TRAFFIC_CONTROL 0
 #define DEFAULT_UPLOAD_LIMIT 0
 #define DEFAULT_DOWNLOAD_LIMIT 0
-#define DEFAULT_DOWNLOAD_IMQ 0
-#define DEFAULT_UPLOAD_IMQ 1
+#define DEFAULT_UPLOAD_IFB 0
 #define DEFAULT_LOG_SYSLOG 0
 #define DEFAULT_SYSLOG_FACILITY LOG_DAEMON
 #define DEFAULT_NDSCTL_SOCK "/tmp/ndsctl.sock"
 #define DEFAULT_INTERNAL_SOCK "/tmp/ndsctl.sock"
-#define DEFAULT_FW_MARK_AUTHENTICATED 0x400
-#define DEFAULT_FW_MARK_TRUSTED 0x200
-#define DEFAULT_FW_MARK_BLOCKED 0x100
-#define DEFAULT_DECONGEST_HTTPD_THREADS 0
-#define DEFAULT_HTTPD_THREAD_THRESHOLD 3
-#define DEFAULT_HTTPD_THREAD_DELAY_MS 200
+#define DEFAULT_FW_MARK_AUTHENTICATED 0x30000
+#define DEFAULT_FW_MARK_TRUSTED 0x20000
+#define DEFAULT_FW_MARK_BLOCKED 0x10000
 /* N.B.: default policies here must be ACCEPT, REJECT, or RETURN
  * In the .conf file, they must be allow, block, or passthrough
  * Mapping between these enforced by parse_empty_ruleset_policy() */
@@ -151,51 +148,48 @@ typedef struct {
 	char *gw_name;			/**< @brief Name of the gateway; e.g. its SSID */
 	char *gw_interface;		/**< @brief Interface we will manage */
 	char *gw_iprange;		/**< @brief IP range on gw_interface we will manage */
-	char *gw_address;		/**< @brief Internal IP address for our web server */
+	char *gw_ip;			/**< @brief Internal IP (v4 or v6) for our web server */
+	char *gw_address;		/**< @brief Internal IP with port for our web server */
 	char *gw_mac;			/**< @brief MAC address of the interface we manage */
 	unsigned int gw_port;		/**< @brief Port the webserver will run on */
-	char *remote_auth_action;	/**< @brief Path for remote auth */
-	char enable_preauth;  		/**< @brief enable pre-authentication support */
-	char *bin_voucher;		/**< @brief enable voucher support */
-	char force_voucher;		/**< @brief force voucher */
+	unsigned int fas_port;		/**< @brief Port the fas server will run on */
+	int fas_secure_enabled;		/**< @brief Enable Secure FAS */
+	char *fas_path;			/**< @brief Path to forward authentication page of FAS */
+	char *fas_key;			/**< @brief AES key for FAS */
+	char *fas_remoteip;		/**< @brief IP addess of a remote FAS */
+	char *fas_remotefqdn;		/**< @brief FQDN of a remote FAS */
+	char *fas_url;			/**< @brief URL of a remote FAS */
+	char *fas_ssl;			/**< @brief SSL provider for FAS */
 	char *webroot;			/**< @brief Directory containing splash pages, etc. */
 	char *splashpage;		/**< @brief Name of main splash page */
-	char *infoskelpage;		/**< @brief Name of info skeleton page */
-	char *imagesdir;		/**< @brief Subdir of webroot containing .png .gif files etc */
-	char *pagesdir;			/**< @brief Subdir of webroot containing other .html files */
+	char *statuspage;		/**< @brief Name of info status page */
 	char *redirectURL;		/**< @brief URL to direct client to after authentication */
 	char *authdir;			/**< @brief Notional relative dir for authentication URL */
 	char *denydir;			/**< @brief Notional relative dir for denial URL */
-	int passwordauth;		/**< @brief boolean, whether to use password authentication */
-	int usernameauth;		/**< @brief boolean, whether to use username authentication */
-	char *username;			/**< @brief Username for username authentication */
-	char *password;			/**< @brief Password for password authentication */
-	int passwordattempts;		/**< @brief Number of attempted password authentications allowed */
-	int clienttimeout;		/**< @brief How many CheckIntervals before an inactive client must be re-authenticated */
-	int clientforceout;		/**< @brief How many CheckIntervals before a client must be re-authenticated */
+	char *preauthdir;		/**< @brief Notional relative dir for preauth URL */
+	int session_timeout;		/**< @brief Minutes of the default session length */
+	int preauth_idle_timeout;	/**< @brief Minutes a preauthenticated client will be kept in the system */
+	int auth_idle_timeout;		/**< @brief Minutes an authenticated client will be kept in the system */
 	int checkinterval;		/**< @brief Period the the client timeout check thread will run, in seconds */
-	int authenticate_immediately;	/**< @brief boolean, whether to auth noninteractively */
 	int set_mss;			/**< @brief boolean, whether to set mss */
 	int mss_value;			/**< @brief int, mss value; <= 0 clamp to pmtu */
 	int traffic_control;		/**< @brief boolean, whether to do tc */
 	int download_limit;		/**< @brief Download limit, kb/s */
 	int upload_limit;		/**< @brief Upload limit, kb/s */
-	int download_imq;		/**< @brief Number of IMQ handling download */
-	int upload_imq;			/**< @brief Number of IMQ handling upload */
+	int upload_ifb;			/**< @brief Number of IFB handling upload */
 	int log_syslog;			/**< @brief boolean, whether to log to syslog */
 	int syslog_facility;		/**< @brief facility to use when using syslog for logging */
-	int decongest_httpd_threads;	/**< @brief boolean, whether to avoid httpd thread congestion */
-	int httpd_thread_threshold; 	/**< @brief number of concurrent httpd threads before trying decongestion */
-	int httpd_thread_delay_ms;	/**< @brief ms delay before starting a httpd thread after threshold */
 	int macmechanism; 		/**< @brief mechanism wrt MAC addrs */
 	t_firewall_ruleset *rulesets;	/**< @brief firewall rules */
 	t_MAC *trustedmaclist;		/**< @brief list of trusted macs */
 	t_MAC *blockedmaclist;		/**< @brief list of blocked macs */
 	t_MAC *allowedmaclist;		/**< @brief list of allowed macs */
-	unsigned int FW_MARK_AUTHENTICATED;	/**< @brief iptables mark for authenticated packets */
-	unsigned int FW_MARK_BLOCKED;	/**< @brief iptables mark for blocked packets */
-	unsigned int FW_MARK_TRUSTED;	/**< @brief iptables mark for trusted packets */
+	unsigned int fw_mark_authenticated;	/**< @brief iptables mark for authenticated packets */
+	unsigned int fw_mark_blocked;	/**< @brief iptables mark for blocked packets */
+	unsigned int fw_mark_trusted;	/**< @brief iptables mark for trusted packets */
 	int ip6;			/**< @brief enable IPv6 */
+	char *binauth;			/**< @brief external authentication program */
+	char *preauth;			/**< @brief external preauthentication program */
 } s_config;
 
 /** @brief Get the current gateway configuration */
@@ -232,6 +226,10 @@ void parse_trusted_mac_list(const char[]);
 void parse_blocked_mac_list(const char[]);
 void parse_allowed_mac_list(const char[]);
 
+int is_blocked_mac(const char *mac);
+int is_allowed_mac(const char *mac);
+int is_trusted_mac(const char *mac);
+
 int add_to_blocked_mac_list(const char possiblemac[]);
 int remove_from_blocked_mac_list(const char possiblemac[]);
 
@@ -245,9 +243,7 @@ int check_ip_format(const char[]);
 int check_mac_format(const char[]);
 
 /** config API, used in commandline.c */
-int set_log_level(int);
-int set_password(const char[]);
-int set_username(const char[]);
+int set_debuglevel(const char[]);
 
 #define LOCK_CONFIG() do { \
 	debug(LOG_DEBUG, "Locking config"); \
