@@ -46,15 +46,24 @@
 #include "auth.h"
 #include "client_list.h"
 #include "fw_iptables.h"
-#include "firewall.h"
 #include "debug.h"
 #include "util.h"
 #include "tc.h"
 
+// iptables v1.4.17
+#define MIN_IPTABLES_VERSION (1 * 10000 + 4 * 100 + 17)
+
 static char *_iptables_compile(const char[], const char[], t_firewall_rule *);
 static int _iptables_append_ruleset(const char[], const char[], const char[]);
 static int _iptables_init_marks(void);
 
+/** Used to mark packets, and characterize client state.  Unmarked packets are considered 'preauthenticated' */
+unsigned int FW_MARK_PREAUTHENTICATED; /**< @brief 0: Actually not used as a packet mark */
+unsigned int FW_MARK_AUTHENTICATED;    /**< @brief The client is authenticated */
+unsigned int FW_MARK_BLOCKED;          /**< @brief The client is blocked */
+unsigned int FW_MARK_TRUSTED;          /**< @brief The client is trusted */
+unsigned int FW_MARK_MASK;             /**< @brief Iptables mask: bitwise or of the others */
+
 extern pthread_mutex_t client_list_mutex;
 extern pthread_mutex_t config_mutex;
 
@@ -74,6 +83,21 @@ static const char* markop = "--set-mark";
 static const char* markmask = "";
 
 
+/** Return a string representing a connection state */
+const char *
+fw_connection_state_as_string(int mark)
+{
+	if (mark == FW_MARK_PREAUTHENTICATED)
+		return "Preauthenticated";
+	if (mark == FW_MARK_AUTHENTICATED)
+		return "Authenticated";
+	if (mark == FW_MARK_TRUSTED)
+		return "Trusted";
+	if (mark == FW_MARK_BLOCKED)
+		return "Blocked";
+	return "ERROR: unrecognized mark";
+}
+
 /** @internal */
 int
 _iptables_init_marks()
@@ -125,23 +149,23 @@ _iptables_check_mark_masking()
 		debug(LOG_DEBUG, "Kernel supports --or-mark.");
 		markop = "--or-mark";
 	} else {
-		debug(LOG_INFO,"Kernel does not support iptables --or-mark.  Using --set-mark instead.");
+		debug(LOG_INFO,"Kernel does not support iptables --or-mark. Using --set-mark instead.");
 		markop = "--set-mark";
 	}
 
 	/* See if kernel supports mark masking */
 	if (0 == iptables_do_command("-t filter -I FORWARD 1 -m mark --mark 0x%x/0x%x -j REJECT", FW_MARK_BLOCKED, FW_MARK_MASK)) {
 		iptables_do_command("-t filter -D FORWARD 1"); /* delete test rule we just inserted */
-		debug(LOG_DEBUG,"Kernel supports mark masking.");
+		debug(LOG_DEBUG, "Kernel supports mark masking.");
 		char *tmp = NULL;
 		safe_asprintf(&tmp,"/0x%x",FW_MARK_MASK);
 		markmask = tmp;
 	} else {
-		debug(LOG_INFO,"Kernel does not support iptables mark masking.  Using empty mask.");
+		debug(LOG_INFO, "Kernel does not support iptables mark masking. Using empty mask.");
 		markmask = "";
 	}
 
-	debug(LOG_INFO,"Iptables mark op \"%s\" and mark mask \"%s\".", markop, markmask);
+	debug(LOG_INFO, "Iptables mark op \"%s\" and mark mask \"%s\".", markop, markmask);
 
 	fw_quiet = 0; /* restore verbosity */
 
@@ -153,8 +177,9 @@ int
 iptables_do_command(const char *format, ...)
 {
 	va_list vlist;
-	char *fmt_cmd = NULL, *cmd = NULL;
+	char *fmt_cmd = NULL;
 	s_config *config;
+	char *iptables;
 	int rc;
 	int i;
 
@@ -164,18 +189,15 @@ iptables_do_command(const char *format, ...)
 
 	config = config_get_config();
 
-	if (config->ip6) {
-		safe_asprintf(&cmd, "ip6tables %s", fmt_cmd);
-	} else {
-		safe_asprintf(&cmd, "iptables %s", fmt_cmd);
-	}
-
-	free(fmt_cmd);
-
-	debug(LOG_DEBUG, "Executing command: %s", cmd);
+	iptables = config->ip6 ? "ip6tables" : "iptables";
 
 	for (i = 0; i < 5; i++) {
-		rc = execute(cmd, fw_quiet);
+		if (fw_quiet) {
+			rc = execute("%s --wait %s > /dev/null 2>&1", iptables, fmt_cmd);
+		} else {
+			rc = execute("%s --wait %s", iptables, fmt_cmd);
+		}
+
 		if (rc == 4) {
 			/* iptables error code 4 indicates a resource problem that might
 			 * be temporary. So we retry to insert the rule a few times. (Mitar) */
@@ -184,11 +206,8 @@ iptables_do_command(const char *format, ...)
 			break;
 		}
 	}
-	if (!fw_quiet && rc != 0) {
-		debug(LOG_ERR, "Nonzero exit status %d from command: %s", rc, cmd);
-	}
 
-	free(cmd);
+	free(fmt_cmd);
 
 	return rc;
 }
@@ -212,23 +231,23 @@ _iptables_compile(const char table[], const char chain[], t_firewall_rule *rule)
 
 	switch (rule->target) {
 	case TARGET_DROP:
-		mode = safe_strdup("DROP");
+		mode = "DROP";
 		break;
 	case TARGET_REJECT:
-		mode = safe_strdup("REJECT");
+		mode = "REJECT";
 		break;
 	case TARGET_ACCEPT:
-		mode = safe_strdup("ACCEPT");
+		mode = "ACCEPT";
 		break;
 	case TARGET_LOG:
-		mode = safe_strdup("LOG");
+		mode = "LOG";
 		break;
 	case TARGET_ULOG:
-		mode = safe_strdup("ULOG");
+		mode = "ULOG";
 		break;
 	}
 
-	snprintf(command, sizeof(command),  "-t %s -A %s ",table, chain);
+	snprintf(command, sizeof(command),  "-t %s -A %s ", table, chain);
 	if (rule->mask != NULL) {
 		snprintf((command + strlen(command)),
 				 (sizeof(command) - strlen(command)),
@@ -253,8 +272,6 @@ _iptables_compile(const char table[], const char chain[], t_firewall_rule *rule)
 			 (sizeof(command) - strlen(command)),
 			 "-j %s", mode);
 
-	free(mode);
-
 	/* XXX The buffer command, an automatic variable, will get cleaned
 	 * off of the stack when we return, so we strdup() it. */
 	return(safe_strdup(command));
@@ -323,16 +340,37 @@ iptables_untrust_mac(const char mac[])
 	return iptables_do_command("-t mangle -D " CHAIN_TRUSTED " -m mac --mac-source %s -j MARK %s 0x%x", mac, markop, FW_MARK_TRUSTED);
 }
 
+int get_iptables_version()
+{
+	char buf[256];
+	int minor;
+	int major;
+	int patch;
+	int rc;
+
+	rc = execute_ret(buf, sizeof(buf), "iptables -V");
+
+	if (rc == 0 && sscanf(buf, "iptables v%d.%d.%d", &major, &minor, &patch) == 3) {
+		return major * 10000 + minor * 100 + patch;
+	} else {
+		return -1;
+	}
+}
+
 /** Initialize the firewall rules.
  */
 int
 iptables_fw_init(void)
 {
 	s_config *config;
-	char * gw_interface = NULL;
-	char * gw_address = NULL;
-	char * gw_iprange = NULL;
+	int iptables_version;
+	char *gw_interface = NULL;
+	char *gw_ip = NULL;
+	char *gw_address = NULL;
+	char *gw_iprange = NULL;
 	int gw_port = 0;
+	char *fas_remoteip = NULL;
+	int fas_port = 0;
 	int traffic_control;
 	int set_mss, mss_value;
 	t_MAC *pt;
@@ -341,9 +379,28 @@ iptables_fw_init(void)
 	int rc = 0;
 	int macmechanism;
 
+	debug(LOG_NOTICE, "Initializing firewall rules");
+
 	LOCK_CONFIG();
 	config = config_get_config();
 	gw_interface = safe_strdup(config->gw_interface); /* must free */
+	
+	/* ip4 vs ip6 differences */
+	const char *ICMP_TYPE;
+	if (config->ip6) {
+		/* ip6 addresses must be in square brackets like [ffcc:e08::1] */
+		safe_asprintf(&gw_ip, "[%s]", config->gw_ip); /* must free */
+		ICMP_TYPE = "icmp6";
+	} else {
+		gw_ip = safe_strdup(config->gw_ip);    /* must free */
+		ICMP_TYPE = "icmp";
+	}
+	
+	if (config->fas_port) {
+		fas_remoteip = safe_strdup(config->fas_remoteip);    /* must free */
+		fas_port = config->fas_port;
+	}
+
 	gw_address = safe_strdup(config->gw_address);    /* must free */
 	gw_iprange = safe_strdup(config->gw_iprange);    /* must free */
 	gw_port = config->gw_port;
@@ -354,11 +411,28 @@ iptables_fw_init(void)
 	set_mss = config->set_mss;
 	mss_value = config->mss_value;
 	traffic_control = config->traffic_control;
-	FW_MARK_BLOCKED = config->FW_MARK_BLOCKED;
-	FW_MARK_TRUSTED = config->FW_MARK_TRUSTED;
-	FW_MARK_AUTHENTICATED = config->FW_MARK_AUTHENTICATED;
+	FW_MARK_BLOCKED = config->fw_mark_blocked;
+	FW_MARK_TRUSTED = config->fw_mark_trusted;
+	FW_MARK_AUTHENTICATED = config->fw_mark_authenticated;
 	UNLOCK_CONFIG();
 
+	iptables_version = get_iptables_version();
+	if (iptables_version < 0) {
+		debug(LOG_ERR, "Cannot get iptables version.");
+		return -1;
+	}
+
+	if (iptables_version < MIN_IPTABLES_VERSION) {
+		debug(LOG_ERR, "Unsupported iptables version v%d.%d.%d, needs at least v%d.%d.%d.",
+			(iptables_version / 10000),
+			(iptables_version % 10000) / 100,
+			(iptables_version % 100),
+			(MIN_IPTABLES_VERSION / 10000),
+			(MIN_IPTABLES_VERSION % 10000) / 100,
+			(MIN_IPTABLES_VERSION % 100)
+		);
+		return -1;
+	}
 
 	/* Set up packet marking methods */
 	rc |= _iptables_init_marks();
@@ -374,6 +448,7 @@ iptables_fw_init(void)
 	/* Create new chains in the mangle table */
 	rc |= iptables_do_command("-t mangle -N " CHAIN_TRUSTED); /* for marking trusted packets */
 	rc |= iptables_do_command("-t mangle -N " CHAIN_BLOCKED); /* for marking blocked packets */
+	rc |= iptables_do_command("-t mangle -N " CHAIN_ALLOWED); /* for marking allowed packets */
 	rc |= iptables_do_command("-t mangle -N " CHAIN_INCOMING); /* for counting incoming packets */
 	rc |= iptables_do_command("-t mangle -N " CHAIN_OUTGOING); /* for marking authenticated packets, and for counting outgoing packets */
 
@@ -400,9 +475,9 @@ iptables_fw_init(void)
 		/* with the MAC_ALLOW mechanism,
 		 * MAC's on the allow list pass;
 		 * everything else is to be marked as blocked */
-		/* So, append at end of chain a rule to mark everything blocked */
+		// So, append at end of chain a rule to mark everything blocked
 		rc |= iptables_do_command("-t mangle -A " CHAIN_BLOCKED " -j MARK %s 0x%x", markop, FW_MARK_BLOCKED);
-		/* But insert at beginning of chain rules to pass allowed MAC's */
+		// But insert at beginning of chain rules to pass allowed MAC's
 		for (; pa != NULL; pa = pa->next) {
 			rc |= iptables_allow_mac(pa->mac);
 		}
@@ -424,33 +499,39 @@ iptables_fw_init(void)
 	/*
 	 *
 	 **************************************
-	 * Set up nat table chains and rules
+	 * Set up nat table chains and rules (ip4 only)
 	 *
 	 */
+	 
+	if (!config->ip6) {
+		/* Create new chains in nat table */
+		rc |= iptables_do_command("-t nat -N " CHAIN_OUTGOING);
 
-	/* Create new chains in nat table */
-	rc |= iptables_do_command("-t nat -N " CHAIN_OUTGOING);
+		/*
+		 * nat PREROUTING chain
+		 */
 
+		// packets coming in on gw_interface jump to CHAIN_OUTGOING
+		rc |= iptables_do_command("-t nat -I PREROUTING -i %s -s %s -j " CHAIN_OUTGOING, gw_interface, gw_iprange);
+		// CHAIN_OUTGOING, packets marked TRUSTED  ACCEPT
+		rc |= iptables_do_command("-t nat -A " CHAIN_OUTGOING " -m mark --mark 0x%x%s -j RETURN", FW_MARK_TRUSTED, markmask);
+		// CHAIN_OUTGOING, packets marked AUTHENTICATED  ACCEPT
+		rc |= iptables_do_command("-t nat -A " CHAIN_OUTGOING " -m mark --mark 0x%x%s -j RETURN", FW_MARK_AUTHENTICATED, markmask);
+		// CHAIN_OUTGOING, append the "preauthenticated-users" ruleset
+		rc |= _iptables_append_ruleset("nat", "preauthenticated-users", CHAIN_OUTGOING);
+
+		// Allow access to remote FAS - CHAIN_OUTGOING and CHAIN_TO_INTERNET packets for remote FAS, ACCEPT
+		if (fas_port && strcmp(fas_remoteip, gw_ip)) {
+			rc |= iptables_do_command("-t nat -A " CHAIN_OUTGOING " -p tcp --destination %s --dport %d -j ACCEPT", fas_remoteip, fas_port);
+		}
+
+		// CHAIN_OUTGOING, packets for tcp port 80, redirect to gw_port on primary address for the iface
+		rc |= iptables_do_command("-t nat -A " CHAIN_OUTGOING " -p tcp --dport 80 -j DNAT --to-destination %s", gw_address);
+		// CHAIN_OUTGOING, other packets ACCEPT
+		rc |= iptables_do_command("-t nat -A " CHAIN_OUTGOING " -j ACCEPT");
+	}
 	/*
-	 * nat PREROUTING chain
-	 */
-
-	/* packets coming in on gw_interface jump to CHAIN_OUTGOING */
-	rc |= iptables_do_command("-t nat -I PREROUTING -i %s -s %s -j " CHAIN_OUTGOING, gw_interface, gw_iprange);
-	/* CHAIN_OUTGOING, packets marked TRUSTED  ACCEPT */
-	rc |= iptables_do_command("-t nat -A " CHAIN_OUTGOING " -m mark --mark 0x%x%s -j ACCEPT", FW_MARK_TRUSTED, markmask);
-	/* CHAIN_OUTGOING, packets marked AUTHENTICATED  ACCEPT */
-	rc |= iptables_do_command("-t nat -A " CHAIN_OUTGOING " -m mark --mark 0x%x%s -j ACCEPT",FW_MARK_AUTHENTICATED, markmask);
-	/* CHAIN_OUTGOING, append the "preauthenticated-users" ruleset */
-	rc |= _iptables_append_ruleset("nat", "preauthenticated-users", CHAIN_OUTGOING);
-
-	/* CHAIN_OUTGOING, packets for tcp port 80, redirect to gw_port on primary address for the iface */
-	rc |= iptables_do_command("-t nat -A " CHAIN_OUTGOING " -p tcp --dport 80 -j DNAT --to-destination %s:%d", gw_address, gw_port);
-	/* CHAIN_OUTGOING, other packets  ACCEPT */
-	rc |= iptables_do_command("-t nat -A " CHAIN_OUTGOING " -j ACCEPT");
-
-	/*
-	 * End of nat table chains and rules
+	 * End of nat table chains and rules (ip4 only)
 	 **************************************
 	 */
 
@@ -461,7 +542,7 @@ iptables_fw_init(void)
 	 *
 	 */
 
-	/* Create new chains in the filter table */
+	// Create new chains in the filter table
 	rc |= iptables_do_command("-t filter -N " CHAIN_TO_INTERNET);
 	rc |= iptables_do_command("-t filter -N " CHAIN_TO_ROUTER);
 	rc |= iptables_do_command("-t filter -N " CHAIN_AUTHENTICATED);
@@ -472,21 +553,26 @@ iptables_fw_init(void)
 	 * filter INPUT chain
 	 */
 
-	/* packets coming in on gw_interface jump to CHAIN_TO_ROUTER */
+	// packets coming in on gw_interface jump to CHAIN_TO_ROUTER
 	rc |= iptables_do_command("-t filter -I INPUT -i %s -s %s -j " CHAIN_TO_ROUTER, gw_interface, gw_iprange);
-	/* CHAIN_TO_ROUTER packets marked BLOCKED  DROP */
+	// CHAIN_TO_ROUTER packets marked BLOCKED DROP
 	rc |= iptables_do_command("-t filter -A " CHAIN_TO_ROUTER " -m mark --mark 0x%x%s -j DROP", FW_MARK_BLOCKED, markmask);
-	/* CHAIN_TO_ROUTER, invalid packets  DROP */
+	// CHAIN_TO_ROUTER, invalid packets DROP
 	rc |= iptables_do_command("-t filter -A " CHAIN_TO_ROUTER " -m conntrack --ctstate INVALID -j DROP");
-	/* CHAIN_TO_ROUTER, related and established packets  ACCEPT */
+	// CHAIN_TO_ROUTER, related and established packets ACCEPT
 	rc |= iptables_do_command("-t filter -A " CHAIN_TO_ROUTER " -m conntrack --ctstate RELATED,ESTABLISHED -j ACCEPT");
-	/* CHAIN_TO_ROUTER, bogus SYN packets  DROP */
-	rc |= iptables_do_command("-t filter -A " CHAIN_TO_ROUTER " -p tcp --tcp-flags SYN SYN \\! --tcp-option 2 -j  DROP");
+	// CHAIN_TO_ROUTER, bogus SYN packets DROP
+	rc |= iptables_do_command("-t filter -A " CHAIN_TO_ROUTER " -p tcp --tcp-flags SYN SYN \\! --tcp-option 2 -j DROP");
 
-	/* CHAIN_TO_ROUTER, packets to HTTP listening on gw_port on router ACCEPT */
+	// CHAIN_TO_ROUTER, packets to HTTP listening on gw_port on router ACCEPT
 	rc |= iptables_do_command("-t filter -A " CHAIN_TO_ROUTER " -p tcp --dport %d -j ACCEPT", gw_port);
 
-	/* CHAIN_TO_ROUTER, packets marked TRUSTED: */
+	// CHAIN_TO_ROUTER, packets to HTTP listening on fas_port on router ACCEPT
+	if (fas_port && !strcmp(fas_remoteip, gw_ip)) {
+		rc |= iptables_do_command("-t filter -A " CHAIN_TO_ROUTER " -p tcp --dport %d -j ACCEPT", fas_port);
+	}
+
+	// CHAIN_TO_ROUTER, packets marked TRUSTED:
 
 	/* if trusted-users-to-router ruleset is empty:
 	 *    use empty ruleset policy
@@ -497,15 +583,15 @@ iptables_fw_init(void)
 		rc |= iptables_do_command("-t filter -A " CHAIN_TO_ROUTER " -m mark --mark 0x%x%s -j %s", FW_MARK_TRUSTED, markmask, get_empty_ruleset_policy("trusted-users-to-router"));
 	} else {
 		rc |= iptables_do_command("-t filter -A " CHAIN_TO_ROUTER " -m mark --mark 0x%x%s -j " CHAIN_TRUSTED_TO_ROUTER, FW_MARK_TRUSTED, markmask);
-		/* CHAIN_TRUSTED_TO_ROUTER, related and established packets  ACCEPT */
+		// CHAIN_TRUSTED_TO_ROUTER, related and established packets ACCEPT
 		rc |= iptables_do_command("-t filter -A " CHAIN_TRUSTED_TO_ROUTER " -m conntrack --ctstate RELATED,ESTABLISHED -j ACCEPT");
-		/* CHAIN_TRUSTED_TO_ROUTER, append the "trusted-users-to-router" ruleset */
+		// CHAIN_TRUSTED_TO_ROUTER, append the "trusted-users-to-router" ruleset
 		rc |= _iptables_append_ruleset("filter", "trusted-users-to-router", CHAIN_TRUSTED_TO_ROUTER);
-		/* CHAIN_TRUSTED_TO_ROUTER, any packets not matching that ruleset  REJECT */
-		rc |= iptables_do_command("-t filter -A " CHAIN_TRUSTED_TO_ROUTER " -j REJECT --reject-with icmp-port-unreachable");
+		// CHAIN_TRUSTED_TO_ROUTER, any packets not matching that ruleset REJECT
+		rc |= iptables_do_command("-t filter -A " CHAIN_TRUSTED_TO_ROUTER " -j REJECT --reject-with %s-port-unreachable", ICMP_TYPE);
 	}
 
-	/* CHAIN_TO_ROUTER, other packets: */
+	// CHAIN_TO_ROUTER, other packets:
 
 	/* if users-to-router ruleset is empty:
 	 *    use empty ruleset policy
@@ -518,7 +604,7 @@ iptables_fw_init(void)
 		/* CHAIN_TO_ROUTER, append the "users-to-router" ruleset */
 		rc |= _iptables_append_ruleset("filter", "users-to-router", CHAIN_TO_ROUTER);
 		/* everything else, REJECT */
-		rc |= iptables_do_command("-t filter -A " CHAIN_TO_ROUTER " -j REJECT --reject-with icmp-port-unreachable");
+		rc |= iptables_do_command("-t filter -A " CHAIN_TO_ROUTER " -j REJECT --reject-with %s-port-unreachable", ICMP_TYPE);
 
 	}
 
@@ -526,13 +612,13 @@ iptables_fw_init(void)
 	 * filter FORWARD chain
 	 */
 
-	/* packets coming in on gw_interface jump to CHAIN_TO_INTERNET */
+	// packets coming in on gw_interface jump to CHAIN_TO_INTERNET
 	rc |= iptables_do_command("-t filter -I FORWARD -i %s -s %s -j " CHAIN_TO_INTERNET, gw_interface, gw_iprange);
-	/* CHAIN_TO_INTERNET packets marked BLOCKED  DROP */
+	// CHAIN_TO_INTERNET packets marked BLOCKED DROP
 	rc |= iptables_do_command("-t filter -A " CHAIN_TO_INTERNET " -m mark --mark 0x%x%s -j DROP", FW_MARK_BLOCKED, markmask);
-	/* CHAIN_TO_INTERNET, invalid packets  DROP */
+	// CHAIN_TO_INTERNET, invalid packets DROP
 	rc |= iptables_do_command("-t filter -A " CHAIN_TO_INTERNET " -m conntrack --ctstate INVALID -j DROP");
-	/* CHAIN_TO_INTERNET, deal with MSS */
+	// CHAIN_TO_INTERNET, deal with MSS
 	if (set_mss) {
 		/* XXX this mangles, so 'should' be done in the mangle POSTROUTING chain.
 		 * However OpenWRT standard S35firewall does it in filter FORWARD,
@@ -544,6 +630,12 @@ iptables_fw_init(void)
 		}
 	}
 
+
+	// Allow access to remote FAS - CHAIN_OUTGOING and CHAIN_TO_INTERNET packets for remote FAS, ACCEPT
+	if (fas_port && strcmp(fas_remoteip, gw_ip)) {
+		rc |= iptables_do_command("-t filter -A " CHAIN_TO_INTERNET " -p tcp --destination %s --dport %d -j ACCEPT", fas_remoteip, fas_port);
+	}
+
 	/* CHAIN_TO_INTERNET, packets marked TRUSTED: */
 
 	/* if trusted-users ruleset is empty:
@@ -555,12 +647,12 @@ iptables_fw_init(void)
 		rc |= iptables_do_command("-t filter -A " CHAIN_TO_INTERNET " -m mark --mark 0x%x%s -j %s", FW_MARK_TRUSTED, markmask, get_empty_ruleset_policy("trusted-users"));
 	} else {
 		rc |= iptables_do_command("-t filter -A " CHAIN_TO_INTERNET " -m mark --mark 0x%x%s -j " CHAIN_TRUSTED, FW_MARK_TRUSTED, markmask);
-		/* CHAIN_TRUSTED, related and established packets  ACCEPT */
+		// CHAIN_TRUSTED, related and established packets ACCEPT
 		rc |= iptables_do_command("-t filter -A " CHAIN_TRUSTED " -m conntrack --ctstate RELATED,ESTABLISHED -j ACCEPT");
-		/* CHAIN_TRUSTED, append the "trusted-users" ruleset */
+		// CHAIN_TRUSTED, append the "trusted-users" ruleset
 		rc |= _iptables_append_ruleset("filter", "trusted-users", CHAIN_TRUSTED);
-		/* CHAIN_TRUSTED, any packets not matching that ruleset  REJECT */
-		rc |= iptables_do_command("-t filter -A " CHAIN_TRUSTED " -j REJECT --reject-with icmp-port-unreachable");
+		// CHAIN_TRUSTED, any packets not matching that ruleset REJECT
+		rc |= iptables_do_command("-t filter -A " CHAIN_TRUSTED " -j REJECT --reject-with %s-port-unreachable", ICMP_TYPE);
 	}
 
 
@@ -575,12 +667,12 @@ iptables_fw_init(void)
 		rc |= iptables_do_command("-t filter -A " CHAIN_TO_INTERNET " -m mark --mark 0x%x%s -j %s", FW_MARK_AUTHENTICATED, markmask, get_empty_ruleset_policy("authenticated-users"));
 	} else {
 		rc |= iptables_do_command("-t filter -A " CHAIN_TO_INTERNET " -m mark --mark 0x%x%s -j " CHAIN_AUTHENTICATED, FW_MARK_AUTHENTICATED, markmask);
-		/* CHAIN_AUTHENTICATED, related and established packets  ACCEPT */
+		// CHAIN_AUTHENTICATED, related and established packets ACCEPT
 		rc |= iptables_do_command("-t filter -A " CHAIN_AUTHENTICATED " -m conntrack --ctstate RELATED,ESTABLISHED -j ACCEPT");
-		/* CHAIN_AUTHENTICATED, append the "authenticated-users" ruleset */
+		// CHAIN_AUTHENTICATED, append the "authenticated-users" ruleset
 		rc |= _iptables_append_ruleset("filter", "authenticated-users", CHAIN_AUTHENTICATED);
-		/* CHAIN_AUTHENTICATED, any packets not matching that ruleset  REJECT */
-		rc |= iptables_do_command("-t filter -A " CHAIN_AUTHENTICATED " -j REJECT --reject-with icmp-port-unreachable");
+		// CHAIN_AUTHENTICATED, any packets not matching that ruleset REJECT
+		rc |= iptables_do_command("-t filter -A " CHAIN_AUTHENTICATED " -j REJECT --reject-with %s-port-unreachable", ICMP_TYPE);
 	}
 
 	/* CHAIN_TO_INTERNET, other packets: */
@@ -591,12 +683,12 @@ iptables_fw_init(void)
 	 *    load and use authenticated-users ruleset
 	 */
 	if (is_empty_ruleset("preauthenticated-users")) {
-		rc |= iptables_do_command("-t filter -A " CHAIN_TO_INTERNET " -j %s ",  get_empty_ruleset_policy("preauthenticated-users"));
+		rc |= iptables_do_command("-t filter -A " CHAIN_TO_INTERNET " -j %s ", get_empty_ruleset_policy("preauthenticated-users"));
 	} else {
 		rc |= _iptables_append_ruleset("filter", "preauthenticated-users", CHAIN_TO_INTERNET);
 	}
-	/* CHAIN_TO_INTERNET, all other packets REJECT */
-	rc |= iptables_do_command("-t filter -A " CHAIN_TO_INTERNET " -j REJECT --reject-with icmp-port-unreachable");
+	// CHAIN_TO_INTERNET, all other packets REJECT
+	rc |= iptables_do_command("-t filter -A " CHAIN_TO_INTERNET " -j REJECT --reject-with %s-port-unreachable", ICMP_TYPE);
 
 	/*
 	 * End of filter table chains and rules
@@ -605,7 +697,9 @@ iptables_fw_init(void)
 
 	free(gw_interface);
 	free(gw_iprange);
+	free(gw_ip);
 	free(gw_address);
+	free(fas_remoteip);
 
 	return rc;
 }
@@ -633,11 +727,7 @@ iptables_fw_destroy(void)
 
 	debug(LOG_DEBUG, "Destroying our iptables entries");
 
-	/*
-	 *
-	 * Everything in the mangle table
-	 *
-	 */
+	/* Everything in the mangle table */
 	debug(LOG_DEBUG, "Destroying chains in the MANGLE table");
 	iptables_fw_destroy_mention("mangle", "PREROUTING", CHAIN_TRUSTED);
 	iptables_fw_destroy_mention("mangle", "PREROUTING", CHAIN_BLOCKED);
@@ -655,22 +745,15 @@ iptables_fw_destroy(void)
 	iptables_do_command("-t mangle -X " CHAIN_OUTGOING);
 	iptables_do_command("-t mangle -X " CHAIN_INCOMING);
 
-	/*
-	 *
-	 * Everything in the nat table
-	 *
-	 */
+	/* Everything in the nat table (ip4 only) */
+	if (!config->ip6) {
+		debug(LOG_DEBUG, "Destroying chains in the NAT table");
+		iptables_fw_destroy_mention("nat", "PREROUTING", CHAIN_OUTGOING);
+		iptables_do_command("-t nat -F " CHAIN_OUTGOING);
+		iptables_do_command("-t nat -X " CHAIN_OUTGOING);
+	}
 
-	debug(LOG_DEBUG, "Destroying chains in the NAT table");
-	iptables_fw_destroy_mention("nat", "PREROUTING", CHAIN_OUTGOING);
-	iptables_do_command("-t nat -F " CHAIN_OUTGOING);
-	iptables_do_command("-t nat -X " CHAIN_OUTGOING);
-
-	/*
-	 *
-	 * Everything in the filter table
-	 *
-	 */
+	/* Everything in the filter table */
 
 	debug(LOG_DEBUG, "Destroying chains in the FILTER table");
 	iptables_fw_destroy_mention("filter", "INPUT", CHAIN_TO_ROUTER);
@@ -686,6 +769,8 @@ iptables_fw_destroy(void)
 	iptables_do_command("-t filter -X " CHAIN_TRUSTED);
 	iptables_do_command("-t filter -X " CHAIN_TRUSTED_TO_ROUTER);
 
+	fw_quiet = 0;
+
 	return 0;
 }
 
@@ -702,6 +787,8 @@ iptables_fw_destroy_mention(
 	const char *mention
 )
 {
+	s_config *config;
+	char *iptables;
 	FILE *p = NULL;
 	char *command = NULL;
 	char *command2 = NULL;
@@ -709,9 +796,11 @@ iptables_fw_destroy_mention(
 	char rulenum[10];
 	int retval = -1;
 
-	debug(LOG_DEBUG, "Checking all mention of %s from %s.%s", mention, table, chain);
+	debug(LOG_DEBUG, "Checking all mention of %s in chain %s of table %s", mention, chain, table);
 
-	safe_asprintf(&command, "iptables -t %s -L %s -n --line-numbers -v", table, chain);
+	config = config_get_config();
+	iptables = config->ip6 ? "ip6tables" : "iptables";
+	safe_asprintf(&command, "%s -t %s -L %s -n --line-numbers -v", iptables, table, chain);
 
 	if ((p = popen(command, "r"))) {
 		/* Skip first 2 lines */
@@ -750,17 +839,14 @@ iptables_fw_destroy_mention(
 /** Insert or delete firewall mangle rules marking a client's packets.
  */
 int
-iptables_fw_access(t_authaction action, t_client *client)
+iptables_fw_authenticate(t_client *client)
 {
 	int rc = 0, download_limit, upload_limit, traffic_control;
 	s_config *config;
-	char *download_imqname, *upload_imqname;
-
-	fw_quiet = 0;
+	char upload_ifbname[16];
 
 	config = config_get_config();
-	safe_asprintf(&download_imqname,"imq%d",config->download_imq); /* must free */
-	safe_asprintf(&upload_imqname,"imq%d",config->upload_imq);  /* must free */
+	sprintf(upload_ifbname, "ifb%d", config->upload_ifb);
 
 	LOCK_CONFIG();
 	traffic_control = config->traffic_control;
@@ -773,35 +859,52 @@ iptables_fw_access(t_authaction action, t_client *client)
 		upload_limit = client->upload_limit;
 	}
 
-	switch(action) {
-	case AUTH_MAKE_AUTHENTICATED:
-		debug(LOG_NOTICE, "Authenticating %s %s", client->ip, client->mac);
-		/* This rule is for marking upload (outgoing) packets, and for upload byte counting */
-		rc |= iptables_do_command("-t mangle -A " CHAIN_OUTGOING " -s %s -m mac --mac-source %s -j MARK %s 0x%x%x", client->ip, client->mac, markop, client->idx + 10, FW_MARK_AUTHENTICATED);
-		rc |= iptables_do_command("-t mangle -A " CHAIN_INCOMING " -d %s -j MARK %s 0x%x%x", client->ip, markop, client->idx + 10, FW_MARK_AUTHENTICATED);
-		/* This rule is just for download (incoming) byte counting, see iptables_fw_counters_update() */
-		rc |= iptables_do_command("-t mangle -A " CHAIN_INCOMING " -d %s -j ACCEPT", client->ip);
-		if (traffic_control) {
-			rc |= tc_attach_client(download_imqname, download_limit, upload_imqname, upload_limit, client->idx, FW_MARK_AUTHENTICATED);
-		}
-		break;
-	case AUTH_MAKE_DEAUTHENTICATED:
-		/* Remove the authentication rules. */
-		debug(LOG_NOTICE, "Deauthenticating %s %s", client->ip, client->mac);
-		rc |= iptables_do_command("-t mangle -D " CHAIN_OUTGOING " -s %s -m mac --mac-source %s -j MARK %s 0x%x%x", client->ip, client->mac, markop, client->idx + 10, FW_MARK_AUTHENTICATED);
-		rc |= iptables_do_command("-t mangle -D " CHAIN_INCOMING " -d %s -j MARK %s 0x%x%x", client->ip, markop, client->idx + 10, FW_MARK_AUTHENTICATED);
-		rc |= iptables_do_command("-t mangle -D " CHAIN_INCOMING " -d %s -j ACCEPT", client->ip);
-		if (traffic_control) {
-			rc |= tc_detach_client(download_imqname, upload_imqname, client->idx);
-		}
-		break;
-	default:
-		rc = -1;
-		break;
+	debug(LOG_NOTICE, "Authenticating %s %s", client->ip, client->mac);
+	/* This rule is for marking upload (outgoing) packets, and for upload byte counting */
+	rc |= iptables_do_command("-t mangle -A " CHAIN_OUTGOING " -s %s -m mac --mac-source %s -j MARK %s 0x%x", client->ip, client->mac, markop, FW_MARK_AUTHENTICATED);
+	rc |= iptables_do_command("-t mangle -A " CHAIN_INCOMING " -d %s -j MARK %s 0x%x", client->ip, markop, FW_MARK_AUTHENTICATED);
+	/* This rule is just for download (incoming) byte counting, see iptables_fw_counters_update() */
+	rc |= iptables_do_command("-t mangle -A " CHAIN_INCOMING " -d %s -j ACCEPT", client->ip);
+
+	if (traffic_control) {
+		rc |= tc_attach_client(config->gw_interface, download_limit, upload_ifbname, upload_limit, client->id, client->ip);
+	}
+
+	return rc;
+}
+
+int
+iptables_fw_deauthenticate(t_client *client)
+{
+	int download_limit, upload_limit, traffic_control;
+	s_config *config;
+	char upload_ifbname[16];
+	int rc = 0;
+
+	config = config_get_config();
+	sprintf(upload_ifbname, "ifb%d", config->upload_ifb);
+
+	LOCK_CONFIG();
+	traffic_control = config->traffic_control;
+	download_limit = config->download_limit;
+	upload_limit = config->upload_limit;
+	UNLOCK_CONFIG();
+
+	if ((client->download_limit > 0) && (client->upload_limit > 0)) {
+		download_limit = client->download_limit;
+		upload_limit = client->upload_limit;
+	}
+
+	/* Remove the authentication rules. */
+	debug(LOG_NOTICE, "Deauthenticating %s %s", client->ip, client->mac);
+	rc |= iptables_do_command("-t mangle -D " CHAIN_OUTGOING " -s %s -m mac --mac-source %s -j MARK %s 0x%x", client->ip, client->mac, markop, FW_MARK_AUTHENTICATED);
+	rc |= iptables_do_command("-t mangle -D " CHAIN_INCOMING " -d %s -j MARK %s 0x%x", client->ip, markop, FW_MARK_AUTHENTICATED);
+	rc |= iptables_do_command("-t mangle -D " CHAIN_INCOMING " -d %s -j ACCEPT", client->ip);
+
+	if (traffic_control) {
+		rc |= tc_detach_client(config->gw_interface, download_limit, upload_ifbname, upload_limit, client->id);
 	}
 
-	free(upload_imqname);
-	free(download_imqname);
 	return rc;
 }
 
@@ -816,7 +919,7 @@ iptables_fw_total_upload()
 	unsigned long long int counter;
 
 	/* Look for outgoing traffic */
-	script =  "iptables -v -n -x -t mangle -L PREROUTING";
+	script = "iptables -v -n -x -t mangle -L PREROUTING";
 	output = popen(script, "r");
 	if (!output) {
 		debug(LOG_ERR, "popen(): %s", strerror(errno));
@@ -827,7 +930,7 @@ iptables_fw_total_upload()
 	while (('\n' != fgetc(output)) && !feof(output)) {}
 	while (('\n' != fgetc(output)) && !feof(output)) {}
 
-	while ( !(feof(output) )) {
+	while (!feof(output)) {
 		rc = fscanf(output, "%*d %llu %s ", &counter, target);
 		if (2 == rc && !strcmp(target,CHAIN_OUTGOING)) {
 			debug(LOG_DEBUG, "Total outgoing Bytes=%llu", counter);
@@ -839,7 +942,7 @@ iptables_fw_total_upload()
 	}
 
 	pclose(output);
-	debug(LOG_ERR, "Can't find target %s in mangle table",CHAIN_OUTGOING);
+	debug(LOG_WARNING, "Can't find target %s in mangle table", CHAIN_OUTGOING);
 	return 0;
 }
 
@@ -854,7 +957,7 @@ iptables_fw_total_download()
 	unsigned long long int counter;
 
 	/* Look for incoming traffic */
-	script =  "iptables -v -n -x -t mangle -L POSTROUTING";
+	script = "iptables -v -n -x -t mangle -L POSTROUTING";
 	output = popen(script, "r");
 	if (!output) {
 		debug(LOG_ERR, "popen(): %s", strerror(errno));
@@ -865,9 +968,9 @@ iptables_fw_total_download()
 	while (('\n' != fgetc(output)) && !feof(output)) {}
 	while (('\n' != fgetc(output)) && !feof(output)) {}
 
-	while ( !(feof(output) )) {
+	while (!feof(output)) {
 		rc = fscanf(output, "%*s %llu %s ", &counter, target);
-		if (2 == rc && !strcmp(target,CHAIN_INCOMING)) {
+		if (2 == rc && !strcmp(target, CHAIN_INCOMING)) {
 			debug(LOG_DEBUG, "Total incoming Bytes=%llu", counter);
 			pclose(output);
 			return counter;
@@ -877,7 +980,7 @@ iptables_fw_total_download()
 	}
 
 	pclose(output);
-	debug(LOG_ERR, "Can't find target %s in mangle table",CHAIN_INCOMING);
+	debug(LOG_WARNING, "Can't find target %s in mangle table", CHAIN_INCOMING);
 	return 0;
 }
 
@@ -886,9 +989,9 @@ int
 iptables_fw_counters_update(void)
 {
 	FILE *output;
-	char *script,
-		ip[INET6_ADDRSTRLEN],
-		target[MAX_BUF];
+	char *script;
+	char ip[INET6_ADDRSTRLEN];
+	char target[MAX_BUF];
 	int rc;
 	int af;
 	s_config *config;
@@ -899,7 +1002,7 @@ iptables_fw_counters_update(void)
 	config = config_get_config();
 	af = config->ip6 ? AF_INET6 : AF_INET;
 
-	/* Look for outgoing traffic */
+	/* Look for outgoing traffic of authenticated clients. */
 	safe_asprintf(&script, "%s %s", "iptables", "-v -n -x -t mangle -L " CHAIN_OUTGOING);
 	output = popen(script, "r");
 	free(script);
@@ -912,11 +1015,11 @@ iptables_fw_counters_update(void)
 	while (('\n' != fgetc(output)) && !feof(output)) {}
 	while (('\n' != fgetc(output)) && !feof(output)) {}
 
-	while ( !(feof(output) )) {
-		rc = fscanf(output, "%*s %llu %s %*s %*s %*s %*s %15[0-9.]", &counter,target,ip);
+	while (!feof(output)) {
+		rc = fscanf(output, "%*s %llu %s %*s %*s %*s %*s %15[0-9.]", &counter, target, ip);
 		/* eat rest of line */
 		while (('\n' != fgetc(output)) && !feof(output)) {}
-		if (3 == rc && !strcmp(target,"MARK")) {
+		if (3 == rc && !strcmp(target, "MARK")) {
 			/* Sanity*/
 			if (!inet_pton(af, ip, &tempaddr)) {
 				debug(LOG_WARNING, "I was supposed to read an IP address but instead got [%s] - ignoring it", ip);
@@ -925,8 +1028,8 @@ iptables_fw_counters_update(void)
 			debug(LOG_DEBUG, "Read outgoing traffic for %s: Bytes=%llu", ip, counter);
 			LOCK_CLIENT_LIST();
 			if ((p1 = client_list_find_by_ip(ip))) {
-				if ((p1->counters.outgoing - p1->counters.outgoing_history) < counter) {
-					p1->counters.outgoing = p1->counters.outgoing_history + counter;
+				if (p1->counters.outgoing < counter) {
+					p1->counters.outgoing = counter;
 					p1->counters.last_updated = time(NULL);
 					debug(LOG_DEBUG, "%s - Updated counter.outgoing to %llu bytes.  Updated last_updated to %d", ip, counter, p1->counters.last_updated);
 				}
@@ -951,11 +1054,11 @@ iptables_fw_counters_update(void)
 	while (('\n' != fgetc(output)) && !feof(output)) {}
 	while (('\n' != fgetc(output)) && !feof(output)) {}
 
-	while ( !(feof(output) )) {
-		rc = fscanf(output, "%*s %llu %s %*s %*s %*s %*s %*s %15[0-9.]", &counter,target,ip);
+	while (!feof(output)) {
+		rc = fscanf(output, "%*s %llu %s %*s %*s %*s %*s %*s %15[0-9.]", &counter, target, ip);
 		/* eat rest of line */
 		while (('\n' != fgetc(output)) && !feof(output)) {}
-		if (3 == rc && !strcmp(target,"ACCEPT")) {
+		if (3 == rc && !strcmp(target, "ACCEPT")) {
 			/* Sanity*/
 			if (!inet_pton(af, ip, &tempaddr)) {
 				debug(LOG_WARNING, "I was supposed to read an IP address but instead got [%s] - ignoring it", ip);
@@ -964,8 +1067,8 @@ iptables_fw_counters_update(void)
 			debug(LOG_DEBUG, "Read incoming traffic for %s: Bytes=%llu", ip, counter);
 			LOCK_CLIENT_LIST();
 			if ((p1 = client_list_find_by_ip(ip))) {
-				if ((p1->counters.incoming - p1->counters.incoming_history) < counter) {
-					p1->counters.incoming = p1->counters.incoming_history + counter;
+				if (p1->counters.incoming < counter) {
+					p1->counters.incoming = counter;
 					debug(LOG_DEBUG, "%s - Updated counter.incoming to %llu bytes", ip, counter);
 				}
 			} else {
