@@ -63,13 +63,10 @@
 #include "util.h"
 #include "conf.h"
 #include "debug.h"
-#include "firewall.h"
 #include "fw_iptables.h"
 
 
-static pthread_mutex_t ghbn_mutex = PTHREAD_MUTEX_INITIALIZER;
-
-/* Defined in gateway.c */
+/* Defined in main.c */
 extern time_t started_time;
 
 /* Defined in clientlist.c */
@@ -79,30 +76,20 @@ extern pthread_mutex_t config_mutex;
 /* Defined in auth.c */
 extern unsigned int authenticated_since_start;
 
-/* Defined in gateway.c */
+/* Defined in main.c */
 extern int created_httpd_threads;
 extern int current_httpd_threads;
 
 
-/** Fork a child and execute a shell command.
- * The parent process waits for the child to return,
- * and returns the child's exit() value.
- * @return Return code of the command
- */
-int
-execute(const char cmd_line[], int quiet)
+static int _execute_ret(char* msg, int msg_len, const char *cmd)
 {
-	int status, retval;
-	pid_t pid, rc;
 	struct sigaction sa, oldsa;
-	const char *new_argv[4];
-	new_argv[0] = "/bin/sh";
-	new_argv[1] = "-c";
-	new_argv[2] = cmd_line;
-	new_argv[3] = NULL;
+	FILE *fp;
+	int rc;
 
-	/* Temporarily get rid of SIGCHLD handler (see gateway.c), until child exits.
-	 * Will handle SIGCHLD here with waitpid() in the parent. */
+	debug(LOG_DEBUG, "Executing command: %s", cmd);
+
+	/* Temporarily get rid of SIGCHLD handler (see main.c), until child exits. */
 	debug(LOG_DEBUG,"Setting default SIGCHLD handler SIG_DFL");
 	sa.sa_handler = SIG_DFL;
 	sigemptyset(&sa.sa_mask);
@@ -111,106 +98,106 @@ execute(const char cmd_line[], int quiet)
 		debug(LOG_ERR, "sigaction() failed to set default SIGCHLD handler: %s", strerror(errno));
 	}
 
-	pid = safe_fork();
-
-	if (pid == 0) {    /* for the child process:         */
-		if (quiet) close(2); /* Close stderr if quiet flag is on */
-		if (execvp("/bin/sh", (char *const *)new_argv) == -1) {    /* execute the command  */
-			debug(LOG_ERR, "execvp(): %s", strerror(errno));
-		} else {
-			debug(LOG_ERR, "execvp() failed");
-		}
-		exit(1);
-
-	} else {        /* for the parent:      */
-		debug(LOG_DEBUG, "Waiting for PID %d to exit", (int)pid);
-		do {
-			rc = waitpid(pid, &status, 0);
-			if (rc == -1) {
-				if (errno == ECHILD) {
-					debug(LOG_DEBUG, "waitpid(): No child exists now. Assuming normal exit for PID %d", (int)pid);
-					retval = 0;
-				} else {
-					debug(LOG_ERR, "Error waiting for child (waitpid() returned -1): %s", strerror(errno));
-					retval = -1;
-				}
-				break;
-			}
-			if (WIFEXITED(status)) {
-				debug(LOG_DEBUG, "Process PID %d exited normally, status %d", (int)rc, WEXITSTATUS(status));
-				retval = (WEXITSTATUS(status));
-			}
-			if (WIFSIGNALED(status)) {
-				debug(LOG_DEBUG, "Process PID %d exited due to signal %d", (int)rc, WTERMSIG(status));
-				retval = -1;
-			}
-		} while (!WIFEXITED(status) && !WIFSIGNALED(status));
-
-		debug(LOG_DEBUG, "Restoring previous SIGCHLD handler");
-		if (sigaction(SIGCHLD, &oldsa, NULL) == -1) {
-			debug(LOG_ERR, "sigaction() failed to restore SIGCHLD handler! Error %s", strerror(errno));
-		}
-
-		return retval;
-	}
-}
-
-struct in_addr *
-wd_gethostbyname(const char name[]) {
-	struct hostent *he;
-	struct in_addr *h_addr, *in_addr_temp;
-
-	/* XXX Calling function is reponsible for free() */
-
-	h_addr = safe_malloc(sizeof(struct in_addr));
-
-	LOCK_GHBN();
-
-	he = gethostbyname(name);
-
-	if (he == NULL) {
-		free(h_addr);
-		UNLOCK_GHBN();
-		return NULL;
+	fp = popen(cmd, "r");
+	if (fp == NULL) {
+		debug(LOG_ERR, "popen(): %s", strerror(errno));
+		rc = -1;
+		goto abort;
 	}
 
-	in_addr_temp = (struct in_addr *)he->h_addr_list[0];
-	h_addr->s_addr = in_addr_temp->s_addr;
+	if (msg && msg_len > 0) {
+		rc = fread(msg, msg_len - 1, 1, fp);
+	}
 
-	UNLOCK_GHBN();
+	rc = pclose(fp);
 
-	return h_addr;
+	if (WIFSIGNALED(rc) != 0) {
+		debug(LOG_WARNING, "Command process exited due to signal %d", WTERMSIG(rc));
+	}
+
+	rc = WEXITSTATUS(rc);
+
+abort:
+
+	/* Restore signal handler */
+	if (sigaction(SIGCHLD, &oldsa, NULL) == -1) {
+		debug(LOG_ERR, "sigaction() failed to restore SIGCHLD handler! Error %s", strerror(errno));
+	}
+
+	return rc;
 }
 
+int execute(const char fmt[], ...)
+{
+	char cmd[QUERYMAXLEN];
+	va_list vlist;
+	int rc;
+
+	va_start(vlist, fmt);
+	rc = vsnprintf(cmd, sizeof(cmd), fmt, vlist);
+	va_end(vlist);
+
+	if (rc < 0 || rc >= sizeof(cmd)) {
+		debug(LOG_ERR, "Format string too small or encoding error.");
+		return -1;
+	}
+
+	return _execute_ret(NULL, 0, cmd);
+}
+
+int execute_ret(char* msg, int msg_len, const char fmt[], ...)
+{
+	char cmd[QUERYMAXLEN];
+	va_list vlist;
+	int rc;
+
+	va_start(vlist, fmt);
+	rc = vsnprintf(cmd, sizeof(cmd), fmt, vlist);
+	va_end(vlist);
+
+	if (rc < 0 || rc >= sizeof(cmd)) {
+		debug(LOG_ERR, "Format string too small or encoding error.");
+		return -1;
+	}
+
+	return _execute_ret(msg, msg_len, cmd);
+}
+
+// Warning: Any client originated portion of the cmd string must be url encoded before calling this function.
+// It may not be desired to url encode the entire cmd string,
+// so it is our responsibility to encode the relevant parts (eg the clients original request url) before calling.
+int execute_ret_url_encoded(char* msg, int msg_len, const char *cmd)
+{
+	return _execute_ret(msg, msg_len, cmd);
+}
+
+
 char *
-get_iface_ip(const char ifname[])
+get_iface_ip(const char ifname[], int ip6)
 {
-	char addrbuf[INET6_ADDRSTRLEN+1];
+	char addrbuf[INET6_ADDRSTRLEN];
 	const struct ifaddrs *cur;
 	struct ifaddrs *addrs;
-	s_config *config;
 
 	if (getifaddrs(&addrs) < 0) {
 		debug(LOG_ERR, "getifaddrs(): %s", strerror(errno));
 		return NULL;
 	}
 
-	config = config_get_config();
-
 	/* Set default address */
-	sprintf(addrbuf, config->ip6 ? "::" : "0.0.0.0");
+	sprintf(addrbuf, ip6 ? "::" : "0.0.0.0");
 
 	/* Iterate all interfaces */
 	cur = addrs;
-	while(cur != NULL) {
-		if ( (cur->ifa_addr != NULL) && (strcmp( cur->ifa_name, ifname ) == 0) ) {
+	while (cur != NULL) {
+		if ((cur->ifa_addr != NULL) && (strcmp( cur->ifa_name, ifname) == 0)) {
 
-			if (config->ip6 && cur->ifa_addr->sa_family == AF_INET6) {
+			if (ip6 && cur->ifa_addr->sa_family == AF_INET6) {
 				inet_ntop(AF_INET6, &((struct sockaddr_in6 *)cur->ifa_addr)->sin6_addr, addrbuf, sizeof(addrbuf));
 				break;
 			}
 
-			if (!config->ip6 && cur->ifa_addr->sa_family == AF_INET) {
+			if (!ip6 && cur->ifa_addr->sa_family == AF_INET) {
 				inet_ntop(AF_INET, &((struct sockaddr_in *)cur->ifa_addr)->sin_addr, addrbuf, sizeof(addrbuf));
 				break;
 			}
@@ -296,62 +283,63 @@ out:
  *  Caller must free.
  */
 char *
-get_ext_iface (void)
+get_ext_iface(void)
 {
 #ifdef __linux__
 	FILE *input;
-	char *device, *gw;
+	char device[16] = {0};
+	char gw[16] = {0};
 	int i = 1;
-	int keep_detecting = 1;
 	pthread_cond_t cond = PTHREAD_COND_INITIALIZER;
 	pthread_mutex_t cond_mutex = PTHREAD_MUTEX_INITIALIZER;
 	struct timespec timeout;
-	device = (char *)malloc(16);
-	gw = (char *)malloc(16);
+
 	debug(LOG_DEBUG, "get_ext_iface(): Autodectecting the external interface from routing table");
-	while(keep_detecting) {
+	for (i = 1; i <= NUM_EXT_INTERFACE_DETECT_RETRY; i += 1) {
 		input = fopen("/proc/net/route", "r");
 		while (!feof(input)) {
-			/* XXX scanf(3) is unsafe, risks overrun */
-			fscanf(input, "%s %s %*s %*s %*s %*s %*s %*s %*s %*s %*s\n", device, gw);
-			if (strcmp(gw, "00000000") == 0) {
-				free(gw);
+			int rc = fscanf(input, "%s %s %*s %*s %*s %*s %*s %*s %*s %*s %*s\n", device, gw);
+			if (rc == 2 && strcmp(gw, "00000000") == 0) {
 				fclose(input);
 				debug(LOG_INFO, "get_ext_iface(): Detected %s as the default interface after try %d", device, i);
-				return device;
+				return strdup(device);
 			}
 		}
 		fclose(input);
 		debug(LOG_ERR, "get_ext_iface(): Failed to detect the external interface after try %d (maybe the interface is not up yet?).  Retry limit: %d", i, NUM_EXT_INTERFACE_DETECT_RETRY);
+
 		/* Sleep for EXT_INTERFACE_DETECT_RETRY_INTERVAL seconds */
 		timeout.tv_sec = time(NULL) + EXT_INTERFACE_DETECT_RETRY_INTERVAL;
 		timeout.tv_nsec = 0;
+
 		/* Mutex must be locked for pthread_cond_timedwait... */
 		pthread_mutex_lock(&cond_mutex);
 		/* Thread safe "sleep" */
 		pthread_cond_timedwait(&cond, &cond_mutex, &timeout);
 		/* No longer needs to be locked */
 		pthread_mutex_unlock(&cond_mutex);
-		//for (i=1; i<=NUM_EXT_INTERFACE_DETECT_RETRY; i++) {
-		if (NUM_EXT_INTERFACE_DETECT_RETRY != 0 && i>NUM_EXT_INTERFACE_DETECT_RETRY) {
-			keep_detecting = 0;
-		}
-		i++;
 	}
+
 	debug(LOG_ERR, "get_ext_iface(): Failed to detect the external interface after %d tries, aborting", i);
 	exit(1);
-	free(device);
-	free(gw);
 #endif
 	return NULL;
 }
 
-/* Malloc's */
 char *
-format_time(unsigned long int secs)
+format_duration(time_t from, time_t to, char buf[64])
 {
-	unsigned int days, hours, minutes, seconds;
-	char *str = NULL;
+	int days, hours, minutes, seconds;
+	long long int secs;
+	const char *neg = "";
+
+	if (from <= to) {
+		secs = to - from;
+	} else {
+		secs = from - to;
+		// Prepend minus sign
+		neg = "-";
+	}
 
 	days = secs / (24 * 60 * 60);
 	secs -= days * (24 * 60 * 60);
@@ -361,40 +349,45 @@ format_time(unsigned long int secs)
 	secs -= minutes * 60;
 	seconds = secs;
 
-	safe_asprintf(&str,"%ud %uh %um %us", days, hours, minutes, seconds);
-	return str;
-}
-
-/* Caller must free. */
-char *
-get_uptime_string()
-{
-	return format_time(time(NULL)-started_time);
-}
-
-/* Custom print format string to file descriptor */
-void
-cprintf( int fd, const char *format, ... ) {
-	char buffer[256];
-	va_list vlist;
-	int rc;
-
-	va_start( vlist, format );
-	rc = vsnprintf( buffer, sizeof(buffer), format, vlist );
-	va_end( vlist );
-
-	if (rc > 0 && rc < sizeof(buffer)) {
-		write(fd, buffer, strlen(buffer));
+	if (days > 0) {
+		snprintf(buf, 64, "%s%dd %dh %dm %ds", neg, days, hours, minutes, seconds);
+	} else if (hours > 0) {
+		snprintf(buf, 64, "%s%dh %dm %ds", neg, hours, minutes, seconds);
+	} else if (minutes > 0) {
+		snprintf(buf, 64, "%s%dm %ds", neg, minutes, seconds);
 	} else {
-		debug(LOG_ERR, "failed to write format string: %s", format);
+		snprintf(buf, 64, "%s%ds", neg, seconds);
 	}
+
+	return buf;
+}
+
+char *
+format_time(time_t time, char buf[64])
+{
+	strftime(buf, 64, "%a %b %d %H:%M:%S %Y", localtime(&time));
+	return buf;
+}
+
+char *
+get_uptime_string(char buf[64])
+{
+	return format_duration(started_time, time(NULL), buf);
+}
+
+int is_addr(const char* addr) {
+	struct sockaddr_in sa;
+	struct sockaddr_in6 sa6;
+
+	return (inet_pton(AF_INET, addr, &sa.sin_addr) == 1) ||
+		(inet_pton(AF_INET6, addr, &sa6.sin6_addr) == 1);
 }
 
 void
-ndsctl_status(int fd)
+ndsctl_status(FILE *fp)
 {
 	char timebuf[32];
-	char *str;
+	char durationbuf[64];
 	s_config *config;
 	t_client *client;
 	int indx;
@@ -406,120 +399,128 @@ ndsctl_status(int fd)
 
 	config = config_get_config();
 
-	cprintf(fd, "==================\nNoDogSplash Status\n====\n");
+	fprintf(fp, "==================\nNoDogSplash Status\n====\n");
 
 	now = time(NULL);
 	uptimesecs = now - started_time;
 
-	cprintf(fd, "Version: " VERSION "\n");
+	fprintf(fp, "Version: " VERSION "\n");
 
-	str = format_time(uptimesecs);
-	cprintf(fd, "Uptime: %s\n", str);
-	free(str);
+	format_duration(started_time, now, durationbuf);
+	fprintf(fp, "Uptime: %s\n", durationbuf);
 
-	cprintf(fd, "Gateway Name: %s\n", config->gw_name);
-	cprintf(fd, "Managed interface: %s\n", config->gw_interface);
-	cprintf(fd, "Managed IP range: %s\n", config->gw_iprange);
-	cprintf(fd, "Server listening: %s:%d\n", config->gw_address, config->gw_port);
+	fprintf(fp, "Gateway Name: %s\n", config->gw_name);
+	fprintf(fp, "Managed interface: %s\n", config->gw_interface);
+	fprintf(fp, "Managed IP range: %s\n", config->gw_iprange);
+	fprintf(fp, "Server listening: http://%s\n", config->gw_address);
 
-	if (config->authenticate_immediately) {
-		cprintf(fd, "Authenticate immediately: yes\n");
+	if (config->binauth) {
+		fprintf(fp, "Binauth Script: %s\n", config->binauth);
 	} else {
-		cprintf(fd, "Splashpage: %s/%s\n", config->webroot, config->splashpage);
+		fprintf(fp, "Binauth: Disabled\n");
 	}
 
+	if (config->preauth) {
+		fprintf(fp, "Preauth Script: %s\n", config->preauth);
+	} else {
+		fprintf(fp, "Preauth: Disabled\n");
+	}
+
+	if (config->fas_port) {
+		fprintf(fp, "FAS: Secure Level %u, URL: %s\n",
+			config->fas_secure_enabled,
+			config->fas_url);
+	} else {
+		fprintf(fp, "FAS: Disabled\n");
+	}
+
+	fprintf(fp, "Client Check Interval: %ds\n", config->checkinterval);
+	fprintf(fp, "Preauth Idle Timeout: %dm\n", config->preauth_idle_timeout);
+	fprintf(fp, "Auth Idle Timeout: %dm\n", config->auth_idle_timeout);
+
 	if (config->redirectURL) {
-		cprintf(fd, "Redirect URL: %s\n", config->redirectURL);
+		fprintf(fp, "Redirect URL: %s\n", config->redirectURL);
 	}
 
-	if (config->passwordauth) {
-		cprintf(fd, "Gateway password: %s\n", config->password);
-	}
-
-	if (config->usernameauth) {
-		cprintf(fd, "Gateway username: %s\n", config->username);
-	}
-
-	cprintf(fd, "Traffic control: %s\n", config->traffic_control ? "yes" : "no");
+	fprintf(fp, "Traffic control: %s\n", config->traffic_control ? "yes" : "no");
 
 	if (config->traffic_control) {
 		if (config->download_limit > 0) {
-			cprintf(fd, "Download rate limit: %d kbit/s\n", config->download_limit);
+			fprintf(fp, "Download rate limit: %d kbit/s\n", config->download_limit);
 		} else {
-			cprintf(fd, "Download rate limit: none\n");
+			fprintf(fp, "Download rate limit: none\n");
 		}
 		if (config->upload_limit > 0) {
-			cprintf(fd, "Upload rate limit: %d kbit/s\n", config->upload_limit);
+			fprintf(fp, "Upload rate limit: %d kbit/s\n", config->upload_limit);
 		} else {
-			cprintf(fd, "Upload rate limit: none\n");
+			fprintf(fp, "Upload rate limit: none\n");
 		}
 	}
 
 	download_bytes = iptables_fw_total_download();
-	cprintf(fd, "Total download: %llu kByte", download_bytes/1000);
-	cprintf(fd, "; avg: %.6g kbit/s\n", ((double) download_bytes) / 125 / uptimesecs);
+	fprintf(fp, "Total download: %llu kByte", download_bytes / 1000);
+	fprintf(fp, "; avg: %.2f kbit/s\n", ((double) download_bytes) / 125 / uptimesecs);
 
 	upload_bytes = iptables_fw_total_upload();
-	cprintf(fd, "Total upload: %llu kByte", upload_bytes/1000);
-	cprintf(fd, "; avg: %.6g kbit/s\n", ((double) upload_bytes) / 125 / uptimesecs);
-	cprintf(fd, "====\n");
-	cprintf(fd, "Client authentications since start: %u\n", authenticated_since_start);
-	cprintf(fd, "Httpd request threads created/current: %d/%d\n", created_httpd_threads, current_httpd_threads);
-
-	if (config->decongest_httpd_threads) {
-		cprintf(fd, "Httpd thread decongest threshold: %d threads\n", config->httpd_thread_threshold);
-		cprintf(fd, "Httpd thread decongest delay: %d ms\n", config->httpd_thread_delay_ms);
-	}
+	fprintf(fp, "Total upload: %llu kByte", upload_bytes / 1000);
+	fprintf(fp, "; avg: %.2f kbit/s\n", ((double) upload_bytes) / 125 / uptimesecs);
+	fprintf(fp, "====\n");
+	fprintf(fp, "Client authentications since start: %u\n", authenticated_since_start);
 
 	/* Update the client's counters so info is current */
 	iptables_fw_counters_update();
 
 	LOCK_CLIENT_LIST();
 
-	cprintf(fd, "Current clients: %d\n", get_client_list_length());
+	fprintf(fp, "Current clients: %d\n", get_client_list_length());
 
 	client = client_get_first_client();
 	if (client) {
-		cprintf(fd, "\n");
+		fprintf(fp, "\n");
 	}
 
 	indx = 0;
 	while (client != NULL) {
-		cprintf(fd, "Client %d\n", indx);
+		fprintf(fp, "Client %d\n", indx);
 
-		cprintf(fd, "  IP: %s MAC: %s\n", client->ip, client->mac);
+		fprintf(fp, "  IP: %s MAC: %s\n", client->ip, client->mac);
 
-		ctime_r(&(client->added_time),timebuf);
-		cprintf(fd, "  Added:   %s", timebuf);
+		format_time(client->counters.last_updated, timebuf);
+		format_duration(client->counters.last_updated, now, durationbuf);
+		fprintf(fp, "  Last Activity: %s (%s ago)\n", timebuf, durationbuf);
 
-		ctime_r(&(client->counters.last_updated),timebuf);
-		cprintf(fd, "  Active:  %s", timebuf);
-
-		str = format_time(client->counters.last_updated - client->added_time);
-		cprintf(fd, "  Active duration: %s\n", str);
-		free(str);
-
-		if (now > client->added_time) {
-			durationsecs = now - client->added_time;
+		if (client->session_start) {
+			format_time(client->session_start, timebuf);
+			format_duration(client->session_start, now, durationbuf);
+			fprintf(fp, "  Session Start: %s (%s ago)\n", timebuf, durationbuf);
 		} else {
-			// prevent divison by 0 later
+			fprintf(fp, "  Session Start: -\n");
+		}
+
+		if (client->session_end) {
+			format_time(client->session_end, timebuf);
+			format_duration(now, client->session_end, durationbuf);
+			fprintf(fp, "  Session End:   %s (%s left)\n", timebuf, durationbuf);
+		} else {
+			fprintf(fp, "  Session End:   -\n");
+		}
+
+		fprintf(fp, "  Token: %s\n", client->token ? client->token : "none");
+
+		fprintf(fp, "  State: %s\n", fw_connection_state_as_string(client->fw_connection_state));
+
+		download_bytes = client->counters.incoming;
+		upload_bytes = client->counters.outgoing;
+		durationsecs = now - client->session_start;
+
+		// prevent divison by 0
+		if (durationsecs < 1) {
 			durationsecs = 1;
 		}
 
-		str = format_time(durationsecs);
-		cprintf(fd, "  Added duration:  %s\n", str);
-		free(str);
-
-		cprintf(fd, "  Token: %s\n", client->token ? client->token : "none");
-
-		cprintf(fd, "  State: %s\n", fw_connection_state_as_string(client->fw_connection_state));
-
-		download_bytes = client->counters.incoming;
-		upload_bytes = client->counters.outgoing;
-
-		cprintf(fd, "  Download: %llu kByte; avg: %.6g kbit/s\n  Upload:   %llu kByte; avg: %.6g kbit/s\n\n",
-				download_bytes/1000, ((double)download_bytes)/125/durationsecs,
-				upload_bytes/1000, ((double)upload_bytes)/125/durationsecs);
+		fprintf(fp, "  Download: %llu kByte; avg: %.2f kbit/s\n  Upload:   %llu kByte; avg: %.2f kbit/s\n\n",
+				download_bytes / 1000, ((double)download_bytes) / 125 / durationsecs,
+				upload_bytes / 1000, ((double)upload_bytes) / 125 / durationsecs);
 
 		indx++;
 		client = client->next;
@@ -527,50 +528,50 @@ ndsctl_status(int fd)
 
 	UNLOCK_CLIENT_LIST();
 
-	cprintf(fd, "====\n");
+	fprintf(fp, "====\n");
 
-	cprintf(fd, "Blocked MAC addresses:");
+	fprintf(fp, "Blocked MAC addresses:");
 
 	if (config->macmechanism == MAC_ALLOW) {
-		cprintf(fd, " N/A\n");
+		fprintf(fp, " N/A\n");
 	} else  if (config->blockedmaclist != NULL) {
-		cprintf(fd, "\n");
+		fprintf(fp, "\n");
 		for (block_mac = config->blockedmaclist; block_mac != NULL; block_mac = block_mac->next) {
-			cprintf(fd, "  %s\n", block_mac->mac);
+			fprintf(fp, "  %s\n", block_mac->mac);
 		}
 	} else {
-		cprintf(fd, " none\n");
+		fprintf(fp, " none\n");
 	}
 
-	cprintf(fd, "Allowed MAC addresses:");
+	fprintf(fp, "Allowed MAC addresses:");
 
 	if (config->macmechanism == MAC_BLOCK) {
-		cprintf(fd, " N/A\n");
+		fprintf(fp, " N/A\n");
 	} else  if (config->allowedmaclist != NULL) {
-		cprintf(fd, "\n");
+		fprintf(fp, "\n");
 		for (allow_mac = config->allowedmaclist; allow_mac != NULL; allow_mac = allow_mac->next) {
-			cprintf(fd, "  %s\n", allow_mac->mac);
+			fprintf(fp, "  %s\n", allow_mac->mac);
 		}
 	} else {
-		cprintf(fd, " none\n");
+		fprintf(fp, " none\n");
 	}
 
-	cprintf(fd, "Trusted MAC addresses:");
+	fprintf(fp, "Trusted MAC addresses:");
 
 	if (config->trustedmaclist != NULL) {
-		cprintf(fd, "\n");
+		fprintf(fp, "\n");
 		for (trust_mac = config->trustedmaclist; trust_mac != NULL; trust_mac = trust_mac->next) {
-			cprintf(fd, "  %s\n", trust_mac->mac);
+			fprintf(fp, "  %s\n", trust_mac->mac);
 		}
 	} else {
-		cprintf(fd, " none\n");
+		fprintf(fp, " none\n");
 	}
 
-	cprintf(fd, "========\n");
+	fprintf(fp, "========\n");
 }
 
 void
-ndsctl_clients(int fd)
+ndsctl_clients(FILE *fp)
 {
 	t_client *client;
 	int indx;
@@ -584,30 +585,35 @@ ndsctl_clients(int fd)
 
 	LOCK_CLIENT_LIST();
 
-	cprintf(fd, "%d\n", get_client_list_length());
+	fprintf(fp, "%d\n", get_client_list_length());
 
 	client = client_get_first_client();
 	if (client) {
-		cprintf(fd, "\n");
+		fprintf(fp, "\n");
 	}
 
 	indx = 0;
 	while (client != NULL) {
-		cprintf(fd, "client_id=%d\n", indx);
-		cprintf(fd, "ip=%s\nmac=%s\n", client->ip, client->mac);
-		cprintf(fd, "added=%lld\n", (long long) client->added_time);
-		cprintf(fd, "active=%lld\n", (long long) client->counters.last_updated);
-		cprintf(fd, "duration=%lu\n", now - client->added_time);
-		cprintf(fd, "token=%s\n", client->token ? client->token : "none");
-		cprintf(fd, "state=%s\n", fw_connection_state_as_string(client->fw_connection_state));
+		fprintf(fp, "client_id=%d\n", indx);
+		fprintf(fp, "ip=%s\nmac=%s\n", client->ip, client->mac);
+		fprintf(fp, "added=%lld\n", (long long) client->session_start);
+		fprintf(fp, "active=%lld\n", (long long) client->counters.last_updated);
+		if (client->session_start) {
+			fprintf(fp, "duration=%lu\n", now - client->session_start);
+		} else {
+			fprintf(fp, "duration=%lu\n", 0ul);
+		}
+		fprintf(fp, "token=%s\n", client->token ? client->token : "none");
+		fprintf(fp, "state=%s\n", fw_connection_state_as_string(client->fw_connection_state));
 
-		durationsecs = now - client->added_time;
+		durationsecs = now - client->session_start;
 		download_bytes = client->counters.incoming;
 		upload_bytes = client->counters.outgoing;
 
-		cprintf(fd, "downloaded=%llu\navg_down_speed=%.6g\nuploaded=%llu\navg_up_speed=%.6g\n\n",
-				download_bytes/1000, ((double)download_bytes)/125/durationsecs,
-				upload_bytes/1000, ((double)upload_bytes)/125/durationsecs);
+		fprintf(fp, "downloaded=%llu\n", download_bytes/1000);
+		fprintf(fp, "avg_down_speed=%.2f\n", ((double)download_bytes) / 125 / durationsecs);
+		fprintf(fp, "uploaded=%llu\n", upload_bytes/1000);
+		fprintf(fp, "avg_up_speed=%.2f\n\n", ((double)upload_bytes) / 125 / durationsecs);
 
 		indx++;
 		client = client->next;
@@ -616,13 +622,40 @@ ndsctl_clients(int fd)
 	UNLOCK_CLIENT_LIST();
 }
 
-void
-ndsctl_json(int fd)
+static void
+ndsctl_json_client(FILE *fp, const t_client *client, time_t now)
+{
+	unsigned long int durationsecs;
+	unsigned long long int download_bytes, upload_bytes;
+
+	fprintf(fp, "\"id\":\"%d\",\n", client->id);
+	fprintf(fp, "\"ip\":\"%s\",\n", client->ip);
+	fprintf(fp, "\"mac\":\"%s\",\n", client->mac);
+	fprintf(fp, "\"added\":\"%lld\",\n", (long long) client->session_start);
+	fprintf(fp, "\"active\":\"%lld\",\n", (long long) client->counters.last_updated);
+	if (client->session_start) {
+		fprintf(fp, "\"duration\":\"%lu\",\n", now - client->session_start);
+	} else {
+		fprintf(fp, "\"duration\":\"%lu\",\n", 0ul);
+	}
+	fprintf(fp, "\"token\":\"%s\",\n", client->token ? client->token : "none");
+	fprintf(fp, "\"state\":\"%s\",\n", fw_connection_state_as_string(client->fw_connection_state));
+
+	durationsecs = now - client->session_start;
+	download_bytes = client->counters.incoming;
+	upload_bytes = client->counters.outgoing;
+
+	fprintf(fp, "\"downloaded\":\"%llu\",\n", download_bytes / 1000);
+	fprintf(fp, "\"avg_down_speed\":\"%.2f\",\n", ((double)download_bytes) / 125 / durationsecs);
+	fprintf(fp, "\"uploaded\":\"%llu\",\n", upload_bytes / 1000);
+	fprintf(fp, "\"avg_up_speed\":\"%.2f\"\n", ((double)upload_bytes)/ 125 / durationsecs);
+}
+
+static void
+ndsctl_json_one(FILE *fp, const char *arg)
 {
 	t_client *client;
-	int indx;
-	unsigned long int now, durationsecs = 0;
-	unsigned long long int download_bytes, upload_bytes;
+	time_t now;
 
 	now = time(NULL);
 
@@ -631,45 +664,65 @@ ndsctl_json(int fd)
 
 	LOCK_CLIENT_LIST();
 
-	cprintf(fd, "{\n\"client_length\": %d,\n", get_client_list_length());
+	client = client_list_find_by_any(arg, arg, arg);
+
+	if (client) {
+		fprintf(fp, "{\n");
+		ndsctl_json_client(fp, client, now);
+		fprintf(fp, "}\n");
+	} else {
+		fprintf(fp, "{}\n");
+	}
+
+	UNLOCK_CLIENT_LIST();
+}
+
+static void
+ndsctl_json_all(FILE *fp)
+{
+	t_client *client;
+	time_t now;
+
+	now = time(NULL);
+
+	/* Update the client's counters so info is current */
+	iptables_fw_counters_update();
+
+	LOCK_CLIENT_LIST();
+
+	fprintf(fp, "{\n\"client_list_length\": \"%d\",\n", get_client_list_length());
 
 	client = client_get_first_client();
-	indx = 0;
 
-	cprintf(fd, "\"clients\":{\n");
+	fprintf(fp, "\"clients\":{\n");
 
 	while (client != NULL) {
-		cprintf(fd, "\"%s\":{\n", client->mac);
-		cprintf(fd, "\"client_id\":%d,\n", indx);
-		cprintf(fd, "\"ip\":\"%s\",\n\"mac\":\"%s\",\n", client->ip, client->mac);
-		cprintf(fd, "\"added\":%lld,\n", (long long) client->added_time);
-		cprintf(fd, "\"active\":%lld,\n", (long long) client->counters.last_updated);
-		cprintf(fd, "\"duration\":%lu,\n", now - client->added_time);
-		cprintf(fd, "\"token\":\"%s\",\n", client->token ? client->token : "none");
-		cprintf(fd, "\"state\":\"%s\",\n", fw_connection_state_as_string(client->fw_connection_state));
+		fprintf(fp, "\"%s\":{\n", client->mac);
+		ndsctl_json_client(fp, client, now);
 
-		durationsecs = now - client->added_time;
-		download_bytes = client->counters.incoming;
-		upload_bytes = client->counters.outgoing;
-
-		cprintf(fd, "\"downloaded\":\"%llu\",\n\"avg_down_speed\":\"%.6g\",\n\"uploaded\":\"%llu\",\n\"avg_up_speed\":\"%.6g\"\n",
-			download_bytes/1000, ((double)download_bytes)/125/durationsecs,
-			upload_bytes/1000, ((double)upload_bytes)/125/durationsecs);
-
-		indx++;
 		client = client->next;
-
-		cprintf(fd, "}");
 		if (client) {
-			cprintf(fd, ",\n");
+			fprintf(fp, "},\n");
+		} else {
+			fprintf(fp, "}\n");
 		}
 	}
 
-	cprintf(fd, "}}" );
+	fprintf(fp, "}\n}\n");
 
 	UNLOCK_CLIENT_LIST();
 }
 
+void
+ndsctl_json(FILE *fp, const char *arg)
+{
+	if (arg && strlen(arg)) {
+		ndsctl_json_one(fp, arg);
+	} else {
+		ndsctl_json_all(fp);
+	}
+}
+
 unsigned short
 rand16(void)
 {
